/*
 * MultiMail offline mail reader
 * miscellaneous routines (global)

 Copyright (c) 1996 Toth Istvan <stoty@vma.bme.hu>
 Copyright (c) 2000 William McBrine <wmcbrine@clark.net>,
                    Peter Karlsson <pk@mds.mdh.se>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */

#include "mmail.h"

// get a little-endian short, return an int
unsigned getshort(const unsigned char *x)
{
	return ((unsigned) x[1] << 8) + (unsigned) x[0];
}

// get a little-endian long
unsigned long getlong(const unsigned char *x)
{
	return ((unsigned long) x[3] << 24) + ((unsigned long) x[2] << 16) +
		((unsigned long) x[1] << 8) + (unsigned long) x[0];
}

// get a big-endian long
unsigned long getblong(const unsigned char *x)
{
	return ((unsigned long) x[0] << 24) + ((unsigned long) x[1] << 16) +
		((unsigned long) x[2] << 8) + (unsigned long) x[3];
}

// put an int into a little-endian short
void putshort(unsigned char *dest, unsigned source)
{
	dest[0] = source & 0xff;
	dest[1] = (source & 0xff00) >> 8;
}

// put a long into a little-endian long
void putlong(unsigned char *dest, unsigned long source)
{
	dest[0] = source & 0xff;
	dest[1] = (source & 0xff00) >> 8;
	dest[2] = (source & 0xff0000) >> 16;
	dest[3] = (source & 0xff000000) >> 24;
}

// put a long into a big-endian long
void putblong(unsigned char *dest, unsigned long source)
{
	dest[0] = (source & 0xff000000) >> 24;
	dest[1] = (source & 0xff0000) >> 16;
	dest[2] = (source & 0xff00) >> 8;
	dest[3] = source & 0xff;
}

// takes off the spaces from the end of a string
char *cropesp(char *st)
{
	char *p;

	for (p = st + strlen(st) - 1; (p > st) && (*p == ' '); p--);
	p[1] = '\0';
	return st;
}

// converts spaces to underline characters
char *unspace(char *source)
{
	for (unsigned c = 0; c < strlen(source); c++)
		if (source[c] == ' ')
			source[c] = '_';
	return source;
}

// allocate and copy a string
char *strdupplus(const char *original)
{
	char *tmp;

	if (original) {
		tmp = new char[strlen(original) + 1];
		strcpy(tmp, original);
	} else
		tmp = 0;

	return tmp;
};

// allocate but do not copy
char *strdupblank(const char *original)
{
	return original ? new char[strlen(original) + 1] : 0;
};

const char *findBaseName(const char *fileName)
{
	const int maxbaselen = 30;
	int c, d;
	static char tmp[maxbaselen + 2];

	for (c = 0; (fileName[c] != '.') && (fileName[c]); c++);

	if (c > maxbaselen)
		c = maxbaselen;

	for (d = 0; d < c; d++)
		tmp[d] = tolower(fileName[d]);
	tmp[d] = '\0';

	return tmp;
};

int getNumExt(const char *fileName)
{
	int retval = -1;
	const char *lastp = strrchr(fileName, '.');

	if (lastp) {
		lastp++;
		if (strlen(lastp) == 3) {
			bool isnum = true;

			for (int x = 0; x < 3; x++)
				isnum = isnum && isdigit(lastp[x]);
			if (isnum)
				retval = atoi(lastp);
		}
	}
	return retval;
}

const char *stripre(const char *subject)
{
	while (!strncasecmp(subject, "re: ", 4))
		subject += 4;
	return subject;
}

// basically the equivalent of "strcasestr()", if there were such a thing
const char *searchstr(const char *source, const char *item)
{
	const char *s;
	char first[3];
	int ilen = strlen(item) - 1;
	bool found = false;

	first[0] = tolower(*item);
	first[1] = toupper(*item);
	first[2] = '\0';

	item++;

	do {
		s = strpbrk(source, first);
		if (s) {
			source = s + 1;
			found = !strncasecmp(source, item, ilen);
		}
	} while (s && !found && *source);

	return found ? s : 0;
}

// Find the address in "Foo <foo@bar.baz>" or "foo@bar.baz (Foo)"
const char *fromAddr(const char *source)
{
	static char tmp[100];
	const char *end = 0, *index = source;

	while (*index) {
		if (*index == '"')
			do
				index++;
			while (*index && (*index != '"'));
		if ((*index == '<') ||
		    ((*index == '(') && (index > source)))
			break;
		if (*index)
			index++;
	}

	if (*index == '<') {
		index++;
		end = strchr(index, '>');
	} else {
		end = index - (*index == '(');
		index = source;
	}

	if (end) {
		int len = end - index;
		if (len > 99)
			len = 99;
		strncpy(tmp, index, len);
		tmp[len] = '\0';
		return tmp;
	}
	return source;
}

// Find the name portion of the address
const char *fromName(const char *source)
{
	static char tmp[100];
	const char *end = 0, *fr = source;

	while (*fr) {
		if (*fr == '"') {
			fr++;
			end = strchr(fr, '"');
			break;
		}
		if (*fr == '(') {
			fr++;
			end = strchr(fr, ')');
			break;
		}
		if ((*fr == '<') && (fr > source)) {
			end = fr - 1;
			fr = source;
			break;
		}
		if (*fr)
			fr++;
	}

	if (end) {
		int len = end - fr;
		if (len > 99)
			len = 99;
		strncpy(tmp, fr, len);
		tmp[len] = '\0';
		return tmp;
	}
	return source;
}

// Should a name be quoted in an address?
bool quoteIt(const char *s)
{
	bool flag = false;

	while (*s) {
		int c = toupper(*s++);
		if (!(((c >= 'A') && (c <= 'Z')) || (c == ' '))) {
			flag = true;
			break;
		}
	}
	return flag;
}

// MIME decoding for header lines (=?iso-8859-1?Q? and =?iso-8859-1?B?)
void headdec(const char *source, const char *cset, char *dest)
{
	// (c) Copyright 1999 Peter Karlsson
	// Modified by William McBrine
	// May be used in any way as long as this copyright is included
	// in the derived work.

	bool isqp, isb64;
	int c, b64buf, b64count;

	if (source)
	    for (isqp = isb64 = false, b64buf = b64count = 0; *source;
	      source++) {
		c = *source;

		if (isqp) {
		    if ('_' == c)		// QP space
			*dest++ = ' ';
		    else
			if (' ' == c) {		// QP end
				*dest++ = ' ';
				isqp = false;
			} else
				if ('?' == c && '=' == source[1]) {
					source++;
					isqp = false;
				} else {
					if ('=' == c) {	// QP escape
						char hex[3] = "00";

						hex[0] = *++source;
						hex[1] = *++source;
						sscanf(hex, "%xd", &c);
					}
					*dest++ = c;
				}
		} else
		    if (isb64) {
			if ('?' == c && '=' == source[1]) {
				source++;
				isb64 = false;
			} else {
				// Update base64 buffer
				static const char *base64 =
				"ABCDEFGHIJKLMNOPQRSTUVWXYZ"	// 0-25
				"abcdefghijklmnopqrstuvwxyz"	// 26-51
				"0123456789+/=";		// 52-63 + pad

				int b64val;
				const char *p = strchr(base64, c);

				b64val = p ? (p - base64) : 0;

				if (64 == b64val)
					b64val = 0;

				b64buf = (b64buf << 6) | b64val;
				b64count++;

				switch (b64count) {
				case 2:
					c = (b64buf & 0xff0) >> 4;
					if (c)
						*dest++ = c;
					b64buf &= 0x0f;
					break;
				case 3:
					c = (b64buf & 0x3fc) >> 2;
					if (c)
						*dest++ = c;
					b64buf &= 0x03;
					break;
				case 4:
					if (b64buf)
						*dest++ = b64buf;
					b64buf = b64count = 0;
				}
			}

		    } else
			if ('=' == c && '?' == source[1]) {
			    int clen = strlen(cset);

			    if (!strncasecmp(source + 2, cset, clen)) {

				char d = toupper(source[3 + clen]);

				if (('Q' == d) || ('B' == d)) {
					source += 4 + clen;

					if ('Q' == d)
						isqp = true;
					else {
						b64buf = b64count = 0;
						isb64 = true;
					}
				} else
					*dest++ = c;
			    } else
				    *dest++ = c;
			} else
				*dest++ = c;
	    }
	*dest = '\0';
}

// length a string will have when encoded per RFC 2047 (pseudo-QP)
int headlen(const unsigned char *in, const char *cset)
{
	int count = 0, clen = strlen(cset);
	bool qp = false;

	while (*in) {
		if (!qp && (*in & 0x80)) {
			count += 5 + clen;
			qp = true;
		}
		if (qp) {
			if (('\"' == *in) ||
			    ((' ' == *in) && ('<' == in[1]))) {
				count += 2;
				qp = false;
			} else
				if (*in & 0x80 || '=' == *in ||
				    '?' == *in || '_' == *in)
					count += 2;
		}

		count++;
		in++;
	}
	if (qp)
		count += 2;

	return count;
}

// write a string to a file in RFC 2047 (pseudo-QP) form
void headenc(const unsigned char *in, const char *cset, FILE *out)
{
	bool qp = false;

	while (*in) {
		if (!qp && (*in & 0x80)) {
			fprintf(out, "=?%s?Q?", cset);
			qp = true;
		}
		if (qp)
			switch (*in) {
			case ' ':
				if ('<' != in[1])
					fputc('_', out);
				else {
					fprintf(out, "?= ");
					qp = false;
				}
				break;
			case '\"':
				fprintf(out, "?=\"");
				qp = false;
				break;
			default:
				if (*in & 0x80 || '=' == *in ||
				    '?' == *in || '_' == *in)
					fprintf(out, "=%02X", *in);
				else
					fputc(*in, out);
			}
		else
			fputc(*in, out);

		in++;
	}
	if (qp)
		fprintf(out, "?=");
}

// decode quoted-printable text in the body of a message
char *qpdecode(const char *source, char *dest, bool qpenc)
{
	while (*source) {
		*dest = *source++;
		if (qpenc && (*dest == '=')) {
			char hex[3];

			hex[0] = *source++;

			if (hex[0] == '\n')
				dest--;
			else {
				int i;

				hex[1] = *source++;
				hex[2] = '\0';
				sscanf(hex, "%xd", &i);
				*dest = i;
			}
		}
		dest++;
	}
	return dest;
}

/*
// write text to a file in Quoted-Printable form (can alter original)
long qpencode(char *src, FILE *dest)
{
	char buf[77], *lastsp = 0;
	long count = 0;
	int col = 0;

	for (; *src; src++) {
		unsigned char c = *src;

		if ('\n' == c) {
			buf[col] = '\0';
			fprintf(dest, "%s\n", buf);
			count += col + 1;
			col = 0;
			lastsp = 0;
		} else {
			if ((c & 0x80) || ('=' == c)) {
				sprintf(buf + col, "=%02X", c);
				col += 3;
			} else {
				if (' ' == c)
					lastsp = src;
				buf[col++] = c;
			}

			if (col > 76) {
				if (!lastsp)
					lastsp = src - 1;
				*lastsp = '\n';
				col -= (src - lastsp);
				src = lastsp - 1;
				buf[col++] = '=';
			}
		}
	}

	if (col) {
		fprintf(dest, "%s\n", buf);
		count += col + 1;
	}

	return count;
}
*/

// write text to a file in Quoted-Printable form
long qpencode(char *src, FILE *dest)
{
	char buf[77];
	char *lastsp = 0;
	long count = 0;
	int col = 0, lastcol = 0;

	for (; *src; src++) {
		unsigned char c = *src;
		bool quotethis = (c & 0x80) || ('=' == c);

		if (((quotethis && (col > 72)) || (col > 74)) &&
		    (src[1] && (src[1] != '\n'))) {

			if (lastcol) {
				col = lastcol + 1;
				src = lastsp;
			} else
				src--;

			buf[col++] = '=';
			c = '\n';
		}


		if ('\n' == c) {
			buf[col] = '\0';
			fprintf(dest, "%s\n", buf);
			count += col + 1;
			col = lastcol = 0;
			lastsp = 0;
		} else {
			if (quotethis) {
				sprintf(buf + col, "=%02X", c);
				col += 3;
			} else {
				if (' ' == c) {
					lastsp = src;
					lastcol = col;
				}
				buf[col++] = c;
			}
		}
	}

	if (col) {
		buf[col] = '\0';
		fprintf(dest, "%s\n", buf);
		count += col + 1;
	}

	return count;
}
